/*
 * File:   communications.c
 * Author: mad_cow
 *
 * Created on April 13, 2018, 11:11 PM
 */

#include <xc.h>
#include "resautage.h"

// Configuration:
// First, the address of this board:
#define OWN_ADDRESS 1
// Then, the first target address, used when the button is pressed short
#define TARGET_ADDRESS_1 2
// A second target address, used when pressing the button for longer
#define TARGET_ADDRESS_2 3

// The protocol used here is quite simple: Every message is one byte.
// The higher 4 bits are the address of the recipient
// The lower 4 bits are a command
// Only one command is used now: Command "1", indicating the button was pressed.

// Just to make life easier...
#define LED LATCbits.LATC5
#define DOIGT PORTAbits.RA2

void main(void)
{
    // Switch oscillator to internl oscillator at 4MHz
    // So we get an Instruction Cycle of 1MHz
    OSCCONbits.IRCF = 5; // Fosc = 4 MHz -> Fcy = 1MHz
    
    // I/O configuration:
    // PC5 is the LED, active low so init to high
    LATCbits.LATC5 = 1;
    TRISCbits.TRISC5 = 0;
    // PA2 is the switch input and gets an internal pullup
    WPUAbits.WPUA2 = 1;
    // PC0 is the externally added pullup for the switch
    // the internal one should work, but doesn't, so I added
    // one on the board, connected to PC0
    LATCbits.LATC0 = 1;
    TRISCbits.TRISC0 = 0;
    // Switch off the ADC function of all ADC pins
    ANSEL = 0;
    ANSELH = 0;
    
    // System Timer
    // Using timer 2 as a compare / match timer here
    // runs for 100 counts -> matches every 100µs
    PR2 = 99;
    // Add a postscaler so it slows down to 1ms
    // (Can't count to 1000 as it's an 8 bit timer)
    T2CONbits.T2OUTPS = 9;
    // Start the timer
    T2CONbits.TMR2ON = 1;
    
    // Interrupt configuration
    // Enable Timer2 interrupt
    PIE1bits.TMR2IE = 1;
    // Enable UART receive interrupt
    PIE1bits.RCIE = 1;
    // Enable interrupts in general
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;
    
    // UART
    // Set data rate: 16 bit divider
    BAUDCONbits.BRG16 = 1;
    // set divider to 26, should lead to 9.6k
    SPBRGH = 0;
    SPBRG = 25; // 9.6k
    // Switch on the transmitter, it will transmit when fed
    // a byte into TXREG
    TXSTAbits.TXEN = 1;
    // Switch on Receiver, will throw an interrupt when a byte comes in
    RCSTAbits.CREN = 1;
    // Enable the UART
    RCSTAbits.SPEN = 1;
    
    // Windows loop: Does nothing, eats up all the time
    while(1);
    return;
}

// Interrupt service routine for all interrupts
void interrupt low_priority distraction(void)
{
    static unsigned int ledTime = 0; // Counts milliseconds up to 500 to blink the LED
    static unsigned int bouncetime = 0; // Counts switch-pressed-time up and down to debounce the switch
    static unsigned int presstime = 0; // Counts debounced switch-pressed-time to distinguish between short and long presses
    static unsigned char ledStatusIn = 0; // status of the LED
    static unsigned int foreign_messages = 0; // counter for messages with the wrong address
    
    if(PIR1bits.TMR2IF) // The interrupt is a timer interrupt
    {
        // Switch debouncing
        if(!DOIGT && bouncetime < 70) bouncetime++;
        else if(bouncetime > 0) bouncetime--;
        if(bouncetime >= 50 && presstime < 5000) presstime++;
        else if(bouncetime <= 20) // button was released
        {
            if(presstime >= 1) // the button was pressed and not read yet
            {
                if(presstime < 1000) // short press -> message to address 1
                {
                    TXREG = (TARGET_ADDRESS_1 << 4 | 1);
                }
                else // long press -> message to address 2
                {
                    TXREG = (TARGET_ADDRESS_2 << 4 | 1);
                }
            }
            presstime = 0;
        }

        // LED blink, only blinks if on
        ledTime++;
        if(ledTime >= 499) // ... doucement!
        {
            if(ledStatusIn == 1) LED ^= 1; // Luire...
            else LED = 1; // Lucas!
            ledTime = 0;
        }

        // Reset the timer interrupt
        PIR1bits.TMR2IF = 0;
    }
    
    if(PIR1bits.RCIF) // The interrupt is a UART receive interrupt
    {
        unsigned char message; // temporary for reading the receive register

        // read RCREG
        message = RCREG;
        
        // check if the message is for us
        if(((message >> 4) & 0x0F) == OWN_ADDRESS) // message is for us
        {
            if((message & 0x0F) == 1) // "button pressed"-command
            {
                if(ledStatusIn) ledStatusIn = 0;
                else ledStatusIn = 1;
            }
            foreign_messages = 0;
        }
        else // message is not for us, forward it
        {
            if(foreign_messages < 500) // this is a crude blocker for endless message loops:
                // if more than 500 messages for other boards come through before one that
                // is addressed for us, we stop forwarding until we get one message addressed
                // to us
            {
                TXREG = message;
                foreign_messages++;
            }
        }
    }
}
